local waitTable = {};
local waitFrame = nil;

local init = false;
local active = true;

-- unit names
local PARTY_UNITS = {"player", "party1", "party2", "party3", "party4"};

-- functions
function MistWeaver_OnLoad(self)
    local _, unitClass = UnitClass("player");
    
    if (unitClass ~= "MONK") then
        return;
    end

    MistWeaverSettings = MistWeaverSettings or {};
    MistWeaverSpells = MistWeaverSpells or SPELLS_DEFAULT;

    self:RegisterEvent("PLAYER_ENTERING_WORLD");

    self:RegisterEvent("PLAYER_REGEN_DISABLED");
    self:RegisterEvent("PLAYER_REGEN_ENABLED");

    self:RegisterEvent("PLAYER_ROLES_ASSIGNED");

    self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
    self:RegisterEvent("UPDATE_SHAPESHIFT_FORM");

    self:RegisterEvent("GROUP_ROSTER_UPDATE");

    self:RegisterEvent("UNIT_HEALTH");
    self:RegisterEvent("UNIT_MAXHEALTH");
    self:RegisterEvent("UNIT_HEAL_PREDICTION");
    self:RegisterEvent("RAID_TARGET_UPDATE");
    self:RegisterEvent("PLAYER_TARGET_CHANGED");
    
    self:RegisterEvent("UNIT_AURA");
    self:RegisterEvent("WAR_MODE_STATUS_UPDATE");
    self:RegisterEvent("PLAYER_FLAGS_CHANGED");

	self:RegisterEvent("PET_BATTLE_OPENING_START");
	self:RegisterEvent("PET_BATTLE_CLOSE");

	self:RegisterEvent("CVAR_UPDATE");

    SLASH_MISTWEAVER1 = "/mw";
    SLASH_MISTWEAVER2 = "/mistweaver";
    SlashCmdList["MISTWEAVER"] = function(msg)
        MistWeaver_SlashCommandHandler(msg);
    end

    MistWeaver_StartDelay(3, MistWeaver_Init);
    MistWeaver_StartDelay(3.4, MistWeaver_FixTextures);
end

function MistWeaver_OnEvent(self, event, ...)
    if (event == "PLAYER_ENTERING_WORLD") then
    end

    if (event == "CVAR_UPDATE") then
        local name, value = ...;
        MistWeaver_CvarUpdate(name, value);
    end

    if (not init) then
        return;
    end

    if (event == "PLAYER_REGEN_ENABLED" or event == "GROUP_ROSTER_UPDATE") then
        MistWeaver_RebindUnitFrames();
    end
    if (event == "PARTY_MEMBERS_CHANGED") then
        MistWeaver_PartyMembersChanged();
    end
    if (event == "PLAYER_FLAGS_CHANGED") then
        -- MistWeaver_StartDelay(2, MistWeaver_AnySpellsChanged);
    end
    if (event == "WAR_MODE_STATUS_UPDATE" or event == "ACTIVE_TALENT_GROUP_CHANGED") then
        -- MistWeaver_StartDelay(2, MistWeaver_AnySpellsChanged);
    end
    if (event == "ACTIVE_TALENT_GROUP_CHANGED" or event == "UPDATE_SHAPESHIFT_FORM") then
        MistWeaver_UpdateActiveState();
    end
    if (event == "PLAYER_ROLES_ASSIGNED") then
        MistWeaver_PlayerRolesAssigned();
    end
    if (event == "UNIT_HEALTH" or event == "UNIT_MAXHEALTH" or event == "UNIT_HEAL_PREDICTION") then
       MistWeaver_UpdateHealPrediction();
    end
    if (event == "UNIT_AURA") then
        -- MistWeaver_UpdatePlayerBuffs();
    end
    if (event == "RAID_TARGET_UPDATE" or event == "PLAYER_TARGET_CHANGED") then
        MistWeaver_UpdateInfoData();
    end

    if (event == "PET_BATTLE_OPENING_START") then
        -- UIFrameFadeOut(MistWeaverFrame, 0.3, MistWeaverFrame:GetAlpha(), 0.0);
    end
    if (event == "PET_BATTLE_CLOSE") then
        -- UIFrameFadeIn(MistWeaverFrame, 0.4, MistWeaverFrame:GetAlpha(), 1.0);
    end
end

function MistWeaver_CvarUpdate(name, value)
    -- mw_debug(name, value)
end

function MistWeaver_SlashCommandHandler(msg)
    if (msg == "reload") then

        ReloadUI();

    elseif (msg == "reset") then

        MistWeaver_Reset();

    elseif (msg == "preset") then

        MistWeaver_Preset();

    elseif (msg == "toggle target") then

        if (InCombatLockdown()) then
            return;
        end
    
        if (MistWeaverSettings.TARGET_HIDDEN == false) then
            MistWeaverSettings.TARGET_HIDDEN = true;
        else
            MistWeaverSettings.TARGET_HIDDEN = false;
        end

        MistWeaver_SetupUnitFramePositions();

    elseif (string.find(msg, "width ") == 1) then

        local str_width = string.sub(msg, 7);
        local width = tonumber(str_width);

        if (width and width >= 50 and width <= 200) then
            MistWeaver_UpdateUnitFrameSize(width, MistWeaverSettings.UNIT_HEIGHT);
        else
            mw_info("|c333399ffwidth|r: 50 - 200");
        end
    
    elseif (string.find(msg, "height ") == 1) then

        local str_height = string.sub(msg, 8);
        local height = tonumber(str_height);

        if (height and height >= 35 and height <= 100) then
            MistWeaver_UpdateUnitFrameSize(MistWeaverSettings.UNIT_WIDTH, height);
        else
            mw_info("|c333399ffheight|r: 35 - 100");
        end
    
    elseif (string.find(msg, "help") == 1) then
        MistWeaver_LogHelp()
    end
end

function MistWeaver_StartDelay(delay, func, ...)
    if(type(delay)~="number" or type(func)~="function") then
        return false;
    end

    if (waitFrame == nil) then
        waitFrame = CreateFrame("Frame","MW_WaitFrame", UIParent);
        waitFrame:SetScript("onUpdate",function (self,elapse)
            local count = #waitTable;
            local i = 1;

            while(i<=count) do
                local waitRecord = tremove(waitTable,i);
                local d = tremove(waitRecord,1);
                local f = tremove(waitRecord,1);
                local p = tremove(waitRecord,1);

                if(d>elapse) then
                    tinsert(waitTable,i,{d-elapse,f,p});
                    i = i + 1;
                else
                    count = count - 1;
                    f(unpack(p));
                end
            end
        end);
    end

    tinsert(waitTable,{delay,func,{...}});

    return true;
end

function MistWeaver_LogHelp()
    mw_info("Chat commands");
    mw_info("Show this info", "/mw help");
    mw_info("Reset all Mistweaver addon settings", "/mw reset");
    mw_info("Reset all Mistweaver addon settings to a default preset", "/mw preset");
    mw_info("Scale the unitframe width (50 - 200)", "/mw width #");
    mw_info("Scale the unitframe height (35 - 100)", "/mw height #");
    mw_info("Show/hide the target unit frame", "/mw toggle target");
end

function MistWeaver_Reset()
    MistWeaverSettings = {};
    MistWeaverSettings.TARGET_HIDDEN = false;

    Mistweaver_ActionButton_ResetSpells(SPELLS_DEFAULT);

    MistWeaver_SetupUnitFramePositions();
    Mistweaver_UpdateActionButtonsAlpha();
    Mistweaver_ActionButton_RebindSpells();
    MistWeaver_UnitFrames_RebindSpells(MistWeaverSpells);
end

function MistWeaver_Preset()
    MistWeaverSettings = {};
    MistWeaverSettings.TARGET_HIDDEN = false;

    Mistweaver_ActionButton_ResetSpells(SPELLS_PRESET);

    MistWeaver_SetupUnitFramePositions();
    Mistweaver_UpdateActionButtonsAlpha();
    Mistweaver_ActionButton_RebindSpells();
    MistWeaver_UnitFrames_RebindSpells(MistWeaverSpells);
end

------------------------------------------------------------------------
---- Init
------------------------------------------------------------------------

function MistWeaver_Init()
    MistWeaver_InitEditmode();
    MistWeaver_InitActionButtons();
    MistWeaver_CreateUnitFrames();

    init = true;

    MistWeaver_UpdateActiveState();
    MistWeaver_PlayerRolesAssigned();

    if (not active) then
        return;
    end

    MistWeaver_RebindUnitFrames();
end

function MistWeaver_UpdateActiveState()

    if (InCombatLockdown()) then
        return;
    end

    -- monk only
    local _, unitClass = UnitClass("player");

    local specialization = GetSpecialization(false, false, GetActiveSpecGroup());

    if (unitClass == "MONK" and specialization == 2) then
        MistWeaverFrame:Show();
        active = true;
    else
        MistWeaverFrame:Hide();
        active = false;
    end
end

------------------------------------------------------------------------
---- OnUpdate
------------------------------------------------------------------------

local UPDATE_WAIT = 0.05;
local wait = 0;
function MistWeaver_OnUpdate(self, elapsed)
    wait = wait + elapsed;

    if (wait > UPDATE_WAIT) then
        wait = 0;
        MistWeaver_DoUpdate(self);
    end
end

function MistWeaver_DoUpdate(self)

    MistWeaver_UnitFrames_DoUpdate();
    MistWeaver_UpdatePower();

    MistWeaver_UpdateUsable();
end
